var NotFound = /*@__PURE__*/(function (Error) {
  function NotFound(route, path) {
    var message = "Unreachable '" + route + "', segment '" + path + "' is not defined";
    Error.call(this, message);
    this.message = message;
  }

  if ( Error ) NotFound.__proto__ = Error;
  NotFound.prototype = Object.create( Error && Error.prototype );
  NotFound.prototype.constructor = NotFound;

  return NotFound;
}(Error));

function buildMatcher(path) {
  var regex;

  var _isSplat;

  var keys = [];
  regex = String(path).replace(/[-/\\^$*+?.()|[\]{}]/g, '\\$&').replace(/\\\((.*?)\\\)/g, '(?:$1)?').replace(/\\?([:*]\w+)/g, function (_, key) {
    keys.push(key.substr(1));

    if (key.charAt() === ':') {
      return '((?!#)[^/]*?)';
    }

    _isSplat = true;
    return '((?!#).*?)';
  });
  regex = new RegExp(("^" + regex + (!_isSplat ? '(?:$|\\/.*?)' : '') + "$"));
  return [keys, regex, _isSplat];
}
var PathMatcher = function PathMatcher(path) {
  var ref = buildMatcher(path);
  var keys = ref[0];
  var regex = ref[1];
  var _isSplat = ref[2];
  return {
    _isSplat: _isSplat,
    match: function (value) {
      var matches = value.match(regex);

      if (matches) {
        return keys.reduce(function (prev, cur, i) {
          prev[cur] = decodeURIComponent(matches[i + 1]);
          return prev;
        }, {});
      }
    }
  };
};

function merge(path, parent) {
  return ("" + (parent && parent !== '/' ? parent : '') + (path || ''));
}
function walk(path, cb) {
  path.split('/').some(function (x) { return cb(("/" + x)); });
}
function sort(routes) {
  routes.sort(function (a, b) {
    var x = a.includes('*') || a.includes(':');
    var y = b.includes('*') || b.includes(':');

    if (x && !y) {
      return a.length > b.length ? -1 : 1;
    }

    return b.length - a.length;
  });
}
function find(path, routes) {
  var leaf = [];
  var out = [];
  var root = routes;
  var splat = false;
  walk(path, function (x) {
    if (splat) {
      return true;
    }

    if (x !== '/') {
      leaf.push(x);
    }

    if (!root[x]) {
      var partial = root.keys.some(function (k) {
        if (root[k].pattern) {
          var matches = root[k].pattern.match(path);

          if (matches) {
            splat = root[k].pattern._isSplat;
            root[k].info.route = root[k].route;
            root[k].info.params = matches;
            root = root[k];
            return true;
          }
        }

        return false;
      });

      if (!partial) {
        throw new NotFound(path, x);
      }
    } else {
      root = root[x];
    }

    var fixedPath = (splat ? path : leaf.join('')) || '/';

    if (root.info && !root.info.route) {
      root.info.route = fixedPath;
    }

    out.push(Object.assign({}, root.info, {
      path: fixedPath
    }));
  });
  return out;
}
function add(path, routes, parent, routeInfo) {
  var fullpath = merge(path, parent);
  var root = routes;
  walk(fullpath, function (x) {
    root.keys = root.keys || [];

    if (!root.keys.includes(x)) {
      root.keys.push(x);
      sort(root.keys);
    }

    root = root[x] || (root[x] = {});
  });
  root.pattern = new PathMatcher(fullpath);
  root.route = fullpath;
  root.info = Object.assign({}, routeInfo);
  return fullpath;
}
function rm(path, routes, parent) {
  var fullpath = merge(path, parent);
  var root = routes;
  var leaf = null;
  var key = null;
  walk(fullpath, function (x) {
    if (!root) {
      leaf = null;
      return true;
    }

    key = x;
    leaf = x === '/' ? routes['/'] : root;
    root = root[x];
  });

  if (!leaf) {
    throw new NotFound(path, key);
  }

  var offset = leaf.keys.indexOf(key);
  leaf.keys.splice(offset, 1);
  delete leaf[key];
}

var Router = function Router() {
  var routes = {};
  var stack = [];
  return {
    mount: function (path, cb) {
      if (path !== '/') {
        stack.push(path);
      }

      cb();
      stack.pop();
    },
    find: function (path) { return find(path, routes); },
    add: function (path, routeInfo) { return add(path, routes, stack.join(''), routeInfo); },
    rm: function (path) { return rm(path, routes, stack.join('')); }
  };
};

export default Router;
